#!/usr/bin/perl -w

package RPM::Source::Editor;
use warnings;
use Carp;
use File::Basename;
use RPM::Source::SpecSection;

our $VERSION=0.26;

sub new {
    my $class = shift;
    my $self = {
	CHANGELOG=> undef,
	PATCHDIR=> '.',
	@_
    };
    bless $self, $class;
    if ($self->{SOURCERPM}) {
	if ($self->{SOURCERPM}!~/.src.rpm$/ && `file $self->{SOURCERPM}`!~/RPM/) {
	    die "$self->{SOURCERPM}: not a rpm file!\n";
	}
	my $dirname=basename($self->{SOURCERPM});
	$dirname=~s/.src.rpm//;
	my $pwd= `pwd`;
	chomp $pwd;
	$self->{TMPDIR}||="$pwd/$dirname.$$";
	$self->{SOURCEDIR}||="$self->{TMPDIR}/SOURCES";
	$self->{SPECDIR}||="$self->{TMPDIR}/SPECS";
	$self->{OUTPUTDIR}||="$self->{TMPDIR}/SRPMS";
	$self->rpm_custom("-i", $self->{SOURCERPM});
	$self->{SPECFILE}=(glob "$self->{TMPDIR}/SPECS/*")[0];
    } elsif ($self->{SPECFILE}) {
	$self->{_SPECFILE_MODE_}=1;
    } else {
	die "$class: new: either SOURCERPM or SPECFILE is required."
    }
    $self->_set_speclist(load_file($self->{SPECFILE}));
    $self->load_macro();
    #print STDERR "done load_macro\n" if $self->{VERBOSE}>2;
    return $self;
}

sub _write_new_spec {
    my $self=shift;
    my $newdir="$self->{TMPDIR}/NEW";
    mkdir $newdir;
    my $specname=basename($self->{SPECFILE});
    my $newspec="$newdir/$specname";
    &write_file($newspec, $self->get_spec());
    $self->{NEWSPECFILE}=$newspec;
    if ($self->{CHANGELOG}) {
	chomp $self->{CHANGELOG};
	run('add_changelog', '--nocheck', '-e',$self->{CHANGELOG},$newspec);
    }
}

sub write_diff {
    my $self=shift;
    $self->_write_new_spec();
    my $diffname=basename($self->{SPECFILE});
    if ($self->{SOURCERPM}) {
	$diffname=basename($self->{SOURCERPM});
	$diffname=~s/\.src\.rpm$//;
    }
    foreach my $i (qw/SPECFILE NEWSPECFILE OUTPUTDIR/) {
	die "Oops!: no $_" unless $self->{$i};
    }
    my $curdir=`pwd`;
    chomp $curdir;
    my $file1=$self->{SPECFILE};
    my $file2=$self->{NEWSPECFILE};
    $file1=~s!^$self->{TMPDIR}/!!;
    $file2=~s!^$self->{TMPDIR}/!!;
    chdir $self->{TMPDIR};
    system ("diff -u $file1 $file2 > $curdir/$self->{OUTPUTDIR}/$diffname.diff");
    chdir $curdir;
}

sub write_rpm {
    my $self=shift;
    $self->_write_new_spec();
    $self->rpm_custom('-bs --nodeps', $self->{NEWSPECFILE});
}

BEGIN {
    our %SECTION=map {$_=>1} qw!
package
description
prep
build
install
clean
files
changelog
pre
post
preun
postun
triggerin
triggerun
triggerpostin
triggerpostun
!
}

sub _parse_speclist {
    my ($rpm, $specptr)=@_;
    my @outpart;
    my $out=[];
    my $TYPE='package';
    foreach (@$specptr) {
	if (/^\s*\%(\w+)\s*/ && $SECTION{$1}) {
	    #my $NEWTYPE=$1;
	    push @outpart, RPM::Source::SpecSection->new(RPM=>$rpm, BODY=>$out);#, TYPE=>$TYPE, PACKAGE=>$PACKAGE);
	    $out=[];
	}
	push @$out, $_;
    }
    push @outpart, RPM::Source::SpecSection->new(RPM=>$rpm, BODY=>$out, TYPE => $TYPE);
    $rpm->{SPECSECTIONS}=\@outpart;
    #foreach (@{$rpm->{SPECSECTIONS}}) {	$_->describe() }
}

sub _reset_speclist {
    my $self=shift;
    my @out;
    foreach my $multiline (@{$self->_get_speclist()}) {
	push @out, split /(?<=\n)/, $multiline;
    }
    $self->_set_speclist(\@out);
}

sub _get_speclist {
    my $self=shift;
    my @out;
    foreach (@{$self->{SPECSECTIONS}}) {
	push @out, @{$_->get_body()};
    }
    return \@out;
}

sub add_section {
    my ($self, $secname, $pkg) =@_;
    $pkg ||= '';
    my $section=$self->get_section($secname,$pkg);
    return $section if $self->get_section($secname,$pkg);
    $self->_raw_add_section($secname, $pkg);
    return $self->get_section($secname,$pkg);
}

sub _raw_add_section {
    my ($self, $secname, $pkg) =@_;
    $pkg ||= '';
    my ($unshiftsec,$pushsec);
    if ($secname eq 'package' || $secname eq 'description') {
	$unshiftsec=$self->get_section('prep');
    } else {
	$unshiftsec=$self->get_section('files');
    }
    $pushsec=$self->get_section('install');
    if ($unshiftsec) {
	$unshiftsec->unshift_body_before_section("
\%$secname $pkg

");
    } elsif ($pushsec) {
	$pushsec->push_body("
\%$secname $pkg

");
    } else {
	Carp::cluck "can't place section $secname :(";
    }

    $self->_reset_speclist();
}

sub is_specfile_mode {
    my $self=shift;
    return $self->{_SPECFILE_MODE_};
}

sub _set_speclist {
    my $self=shift;
    my $ptr=shift;
    die "not an array pointer!" if ! ref $ptr eq 'ARRAY';
    $self->_parse_speclist($ptr);
}

sub get_spec {
    my $self=shift;
    return join('',@{$self->_get_speclist()})
}

sub get_section {
    my $self=shift;
    my $type=shift;
    my $pkg=shift;
    $pkg='' if (!defined $pkg);
    foreach my $section (@{$self->{SPECSECTIONS}}) {
	return $section if $section->get_type() eq $type and $self->are_package_names_equal($section->get_package(),$pkg);
    }
    return;
}

sub are_package_names_equal {
    my ($self, $name1, $name2)=@_;
    $name1=_package_name_strip_options($name1);
    $name2=_package_name_strip_options($name2);
    $is_name1_n = substr($name1,0,2) eq '-n';
    $is_name2_n = substr($name2,0,2) eq '-n';
    if ($is_name1_n or $is_name2_n) {
	$name1='-n %{name}-'.$name1 unless $is_name1_n;
	$name2='-n %{name}-'.$name2 unless $is_name2_n;
	# hack around comparison -n %{name} and ''
	$name1=~s/-$//;
	$name2=~s/-$//;
    }
    $name1=&macro_subst($self->{MACRO},$name1);
    $name2=&macro_subst($self->{MACRO},$name2);
    #print STDERR "are_package_names_equal: [$name1] [$name2]\n";
    # TODO: -n handler
    return $name1 eq $name2;
}

sub _package_name_strip_options {
    my $name=shift;
    $name=~s/^\s*//;
    $name=~s/\s*$//;
    while($name=~s/^-(?:l)\s+\w+\s*//) {};
    return $name;
}

sub disable_package {
    my $self=shift;
    my $name=shift;
    $name=~s/^\s*//;
    $name=~s/\s*$//;
    my $keyname=$name;
    $keyname=~s/^.+\s+//;
    $keyname=~s/-/_/;
    unless ($keyname) {
	$keyname='strange_package';
	warn "disable_package: strange name $name";
    }
    my $is_applied=0;
    unless ($name) {
	Carp::carp "main package can't be disabled!\n";
	return;
    }
    $self->{SPECSECTIONS}->[0]->unshift_body('%def_without '.$keyname."\n");
    for (my $i=0; $i<@{$self->{SPECSECTIONS}}; $i++) {
	my $section = $self->{SPECSECTIONS}->[$i];
	if ($self->are_package_names_equal($section->get_package(), $name)) {
	    $section->unshift_body_before_section('%if_with '.$keyname."\n");
	    $section->push_body('%endif #'.$keyname."\n");
	    $is_applied = 1;
	}
    }
    $is_applied || $self->_warn_notapplied("not disabled [$name] - not found");

    return;
}

sub raw_rename_section {
    my $self=shift;
    my $oldname=shift;
    my $newname=shift;
    foreach my $section (@{$self->{SPECSECTIONS}}) {
	$section->{PACKAGE}=$newname if ($section->get_body()->[0]=~s/(\s*\%\w+\s+)($oldname)\s*/$1$newname\n/);
    }
    return;
}

sub get_sections {
    my $self=shift;
    #return @{$self->{SPECSECTIONS}};
    return $self->{SPECSECTIONS};
}

sub set_changelog {
    my $self=shift;
    return $self->{CHANGELOG}=shift();
}

sub append_changelog {
    my $self=shift;
    return $self->{CHANGELOG}.=shift();
}

sub rpm_custom {
    my $self=shift;
    my $cmd =qq{rpm --define "_specdir $self->{SPECDIR}" --define "_sourcedir $self->{SOURCEDIR}" --define "_srcrpmdir $self->{OUTPUTDIR}" @_};
    &run($cmd);
}

sub DESTROY {
    my $self=shift;
    $self->cleanup();
}

sub cleanup {
    my $self=shift;
    my $tmpdir=$self->{TMPDIR};
    run("rm -rf $tmpdir") if $tmpdir;
}

sub copy_to_sources {
    my $self=shift;
    return if $self->is_specfile_mode();
    foreach my $file (@_) {
	if (-e $file) {
	    run ("cp -f $file $self->{SOURCEDIR}/");
	} else {
	    run ("cp -f $self->{PATCHDIR}/$file $self->{SOURCEDIR}/");
	}
    }
}

sub add_patch {
    my $self=shift;
    my $name=shift;
    my %param = (
	NAME=> $name,
	STRIP => 1,
	NUMBER => 33,
	@_
	);
    # find first unused patch position
    while ($self->{PATCHES}->[$param{NUMBER}]) {$param{NUMBER}++;}
    # register patch
    $self->{PATCHES}->[$param{NUMBER}]=$param{NAME};
    # add patch to spec
    $self->get_section('prep')->push_body('%patch'.$param{NUMBER}.' -p'.$param{STRIP}."\n");
    $self->get_section('package','')->push_body('Patch'.$param{NUMBER}.': '.$param{NAME}."\n");
    $self->copy_to_sources($self->{PATCHDIR}.'/'.$param{NAME});
}

sub add_source {
    my $self=shift;
    my $name=shift;
    my %param = (
	NAME=> $name,
	NUMBER => 44,
	@_
	);
    # find first unused patch position
    while ($self->{SOURCES}->[$param{NUMBER}]) {$param{NUMBER}++;}
    # register patch
    $self->{SOURCES}->[$param{NUMBER}]=$param{NAME};
    # add to spec
    $self->get_section('package','')->push_body('Source'.$param{NUMBER}.': '.$param{NAME}."\n");
    $self->copy_to_sources($self->{PATCHDIR}.'/'.$param{NAME});
    return $param{NUMBER};
}

sub run {
    my (@cmd)=@_;
    system(@cmd) && Carp::confess ("@cmd failed: $?");
}

sub load_file {
    my ($name)=@_;
    open FN, $name || die "can't open $name: $!";
    my @result = <FN>;
    close FN, $name || die "can't open $name: $!";
    return \@result;
}

sub write_array_to_file {
    my ($name, $ptr)=@_;
    open FN, '>', $name || die "can't open $name: $!";
    print FN @$ptr;
    close FN, $name || die "can't open $name: $!";
}

sub write_file {
    my ($name, $str)=@_;
    open FN, '>', $name || die "can't open $name: $!";
    print FN $str;
    close FN, $name || die "can't open $name: $!";
}

sub load_macro {
    my $self=shift;
    print STDERR "entering load_macro\n" if $self->{VERBOSE}>2;
    my $specfile = $self->{SPECFILE};
    my %macro;
    $macro{'nil'}='';
    $macro{'homedir'}=$ENV{'HOME'};
    $macro{'_topsrcdir'}='%_topdir';
    $macro{'_sourcedir'}="%_topsrcdir/SOURCES";
    $macro{'_specdir'}="%_topsrcdir/SPECS";
    $macro{'_srcrpmdir'}="%_topdir/SRPMS";
    if (-e $ENV{'HOME'}.'/.rpmmacros') {
	open (RPMMACROS, $ENV{'HOME'}.'/.rpmmacros') || die "can't open .rpmmacros:$!";
	while (<RPMMACROS>) {
	    if (/^[%]([\w_\/.\d]+)\s+(.*\S)\s*$/) {
		$macro{$1}=$2 ;
		&macro_subst_inplace(\%macro,\$macro{$1});
	    }
	}
	close (RPMMACROS);
    }
    $macro{'_topdir'}||="$ENV{'HOME'}/RPM";
    
    my @source;
    my @patch;
    my @setup_n;
    my @is_setup;
    
    open (SPECFILE, $specfile) || die "can't open spec $specfile: $!";
    while (<SPECFILE>) {
	chomp;
	# to avoid recursion Name: %name; old hack is && not $macro{'version'};
	$macro{'name'}=&macro_subst(\%macro, $1) if /^Name\s*:\s*(.*\S)\s*$/;
	$macro{'version'}=&macro_subst(\%macro, $1) if /^Version\s*:\s*(.*\S)\s*$/;
	$macro{'release'}=&macro_subst(\%macro, $1) if /^Release\s*:\s*(.*\S)\s*$/;

	$macro{$1}=$2 if /^\s*[%]define\s+([\w_\d]+)\s+(\S+)\s*$/;
	# not needed there
	$source[$1 ? $1 : 0]=$3 if /^Source(\d*)\s*:\s*(.+\/)?([^\/]+)\s*$/;
	$patch[$1 ? $1 : 0]=$3 if /^Patch(\d*)\s*:\s*(.+\/)?([^\/]+)\s*$/;
	$setup_n[$1 ? $1 : 0] =$3 if /^[%]setup(\d*)\s*(.+)?-n\s+(\S+)/;
	$is_setup[$1 ? $1 : 0] =1 if /^[%]setup(\d*)/;
    }
    close (SPECFILE);

    my @rawsource=@source;
    my @rawpatch=@patch;
    foreach my $dep (grep {$_} @source, @patch) {
	print STDERR "loop3 $dep\n" if $self->{VERBOSE}>2;
	&macro_subst_inplace(\%macro,\$dep);
    }
    #print join("\n",@require),"\n";
    #print join("\n",keys %macro),"\n";

    foreach my $key  
	(qw/
     _sourcedir
     _specdir
     _srcrpmdir
     name
     version
     release
     /) {
	    if (!$macro{$key}) {
		print STDERR "WARNING: undefined $key!\n";
	    }
	    &macro_subst_inplace(\%macro,\$macro{$key});
    }
    $self->{MACRO}=\%macro;
    $self->{SOURCES}=\@source;
    $self->{PATCHES}=\@patch;
    $self->{RAWSOURCES}=\@rawsource;
    $self->{RAWPATCHES}=\@rawpatch;
    $self->{SETUP_N}=\@setup_n;
    $self->{IS_SETUP}=\@is_setup;
}

sub macro_subst {
    my $macroptr=shift;
    my $arg=shift;
    &macro_subst_inplace($macroptr, \$arg);
    #print STDERR "macro subst: $arg\n";
    return $arg;
}


sub macro_subst_inplace {
    my $macroptr=shift;
    my $argptr=shift;
    #print STDERR "subst: $$argptr\n";
    Carp::cluck "undef" unless defined $$argptr;
    while ($$argptr=~/\%\{?([\w_\d]+)\}?/ and defined $macroptr->{$1}) {
	#print STDERR ":: $$argptr\n";
	my $oldval=$$argptr;
	$$argptr=~s/\%\{?([\w_\d]+)\}?/$macroptr->{$1}/;
	if ($oldval eq $$argptr) {
	    warn "macro substitution recursion for ".$oldval."\n";
	    last;
	}
    }
}

1;

=head1	NAME

Test::Repocop::SQL -   for repocop

=head1	DESCRIPTION

TODO

=head1	AUTHOR

Written by Igor Vlasenko <viy@altlinux.org>.

=head1	COPYING

Copyright (c) 2008 Igor Vlasenko, ALT Linux Team.

This is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

=cut

